#define com_maroontress_libcgi_request_Pair_IMPORT
#define com_maroontress_libcgi_util_Tub_IMPORT
#define std_io_InputTube_IMPORT

#include <stdlib.h>
#include <ctype.h>
#include <com/maroontress/libcgi/request/Pair.h>
#include <com/maroontress/libcgi/util/Tub.h>
#include <std/io/InputTube.h>

struct Pair {
    char *key;
    char *value;
};

struct Pair *
Pair_create(void)
{
    struct Pair *p;

    if ((p = (struct Pair *)malloc(sizeof(*p))) == NULL)
	return NULL;
    p->key = NULL;
    p->value = NULL;
    return p;
}

void
Pair_free(struct Pair *p)
{
    if (p == NULL)
	return;
    free(p->key);
    free(p->value);
    free(p);
}

void
Pair_clear(struct Pair *p)
{
    free(p->key);
    p->key = NULL;
    free(p->value);
    p->value = NULL;
}

void
Pair_setKey(struct Pair *p, char *key)
{
    free(p->key);
    p->key = key;
}

void
Pair_setValue(struct Pair *p, char *value)
{
    free(p->value);
    p->value = value;
}

char *
Pair_getKey(struct Pair *p)
{
    return p->key;
}

char *
Pair_getValue(struct Pair *p)
{
    return p->value;
}

static char *
readKey(union InputTube in, struct Tub *tub)
{
    int c;

    Tub_clear(tub);
    while ((c = InputTube_readByte(in)) != InputTube_EOS
	   && c != '='
	   && Tub_writeByte(tub, c) == 0) {
	continue;
    }
    if (c != '=' || Tub_hasError(tub) || Tub_writeByte(tub, 0) < 0)
	return NULL;
    return Tub_createByteArray(tub);
}

static unsigned int
getXDigit(int c)
{
    c = tolower(c);
    if (c >= '0' && c <= '9')
	return c - '0';
    return c - 'a' + 10;
}

static void
decodeURL(unsigned char *p)
{
    unsigned char *q;

    for (q = p; *p; ++p) {
        switch (*p) {
        case '%':
            if (isxdigit(p[1]) && isxdigit(p[2])) {
		*q = (getXDigit(p[1]) << 4) | getXDigit(p[2]);
                p += 2;
            }
            else {
                *q = *p;
            }
            break;
        case '+':
            *q = ' ';
            break;
        default:
            *q = *p;
            break;
        }
        ++q;
    }
    *q = 0;
}

static char *
readValue(union InputTube in, struct Tub *tub)
{
    int c;
    char *str;

    Tub_clear(tub);
    while ((c = InputTube_readByte(in)) != InputTube_EOS
	   && c != '&'
	   && Tub_writeByte(tub, c) == 0) {
	continue;
    }
    if (Tub_hasError(tub) || Tub_writeByte(tub, 0) < 0)
	return NULL;
    if ((str = Tub_createByteArray(tub)) == NULL)
	return NULL;
    decodeURL(str);
    return str;
}

int
Pair_read(struct Pair *pair, union InputTube in, struct Tub *tub)
{
    char *str;

    Pair_clear(pair);
    if ((str = readKey(in, tub)) == NULL) {
	return (InputTube_hasError(in)) ? -1 : 1;
    }
    Pair_setKey(pair, str);
    if ((str = readValue(in, tub)) == NULL) {
	return -1;
    }
    Pair_setValue(pair, str);
    return 0;
}
